@@ -33,6 +33,10 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
+#ifdef HAVE_SYS_PRCTL_H
+#include <sys/prctl.h>
+#endif
+
 #include "global.h"
 #include "ibusimpl.h"
 #include "server.h"
@@ -43,12 +47,13 @@ static gboolean xim = FALSE;
 static gboolean replace = FALSE;
 static gboolean restart = FALSE;
 static gchar *panel = "default";
-static gchar *panel_extension = "default";
+static gchar *emoji_extension = "default";
 static gchar *config = "default";
 static gchar *desktop = "gnome";
 
 static gchar *panel_extension_disable_users[] = {
     "gdm",
+    "gnome-initial-setup",
     "liveuser"
 };
 
@@ -67,7 +72,7 @@ static const GOptionEntry entries[] =
     { "xim",       'x', 0, G_OPTION_ARG_NONE,   &xim,       "execute ibus XIM server.", NULL },
     { "desktop",   'n', 0, G_OPTION_ARG_STRING, &desktop,   "specify the name of desktop session. [default=gnome]", "name" },
     { "panel",     'p', 0, G_OPTION_ARG_STRING, &panel,     "specify the cmdline of panel program. pass 'disable' not to start a panel program.", "cmdline" },
-    { "panel-extension", 'E', 0, G_OPTION_ARG_STRING, &panel_extension, "specify the cmdline of panel extension program. pass 'disable' not to start an extension program.", "cmdline" },
+    { "emoji-extension", 'E', 0, G_OPTION_ARG_STRING, &emoji_extension, "specify the cmdline of emoji extension program. pass 'disable' not to start an extension program.", "cmdline" },
     { "config",    'c', 0, G_OPTION_ARG_STRING, &config,    "specify the cmdline of config program. pass 'disable' not to start a config program.", "cmdline" },
     { "address",   'a', 0, G_OPTION_ARG_STRING, &g_address,   "specify the address of ibus daemon.", "address" },
     { "replace",   'r', 0, G_OPTION_ARG_NONE,   &replace,   "if there is an old ibus-daemon is running, it will be replaced.", NULL },
@@ -160,17 +165,14 @@ daemon (gint nochdir, gint noclose)
 }
 #endif
 
-/*
- * _sig_usr2_handler:
- * @sig: the signal number, which is usually SIGUSR2.
- *
- * A signal handler for SIGUSR2 signal. Dump a summary of memory usage to stderr.
- */
+#ifdef HAVE_SYS_PRCTL_H
 static void
-_sig_usr2_handler (int sig)
+_sig_usr1_handler (int sig)
 {
-    g_mem_profile ();
+    g_warning ("The parent process died.");
+    bus_server_quit (FALSE);
 }
+#endif
 
 gint
 main (gint argc, gchar **argv)
@@ -196,8 +198,7 @@ main (gint argc, gchar **argv)
     }
 
     if (g_mempro) {
-        g_mem_set_vtable (glib_mem_profiler_table);
-        signal (SIGUSR2, _sig_usr2_handler);
+        g_warning ("--mem-profile no longer works with the GLib 2.46 or later");
     }
 
     /* check uid */
@@ -213,7 +214,7 @@ main (gint argc, gchar **argv)
     /* daemonize process */
     if (daemonize) {
         if (daemon (1, 0) != 0) {
-            g_printerr ("Can not daemonize ibus.\n");
+            g_printerr ("Cannot daemonize ibus.\n");
             exit (-1);
         }
     }
@@ -245,7 +246,7 @@ main (gint argc, gchar **argv)
     bus_server_init ();
     for (i = 0; i < G_N_ELEMENTS(panel_extension_disable_users); i++) {
         if (!g_strcmp0 (username, panel_extension_disable_users[i]) != 0) {
-            panel_extension = "disable";
+            emoji_extension = "disable";
             break;
         }
     }
@@ -286,21 +287,21 @@ main (gint argc, gchar **argv)
     }
 
 #ifdef EMOJI_DICT
-    if (g_strcmp0 (panel_extension, "default") == 0) {
+    if (g_strcmp0 (emoji_extension, "default") == 0) {
         BusComponent *component;
         component = bus_ibus_impl_lookup_component_by_name (
                 BUS_DEFAULT_IBUS, IBUS_SERVICE_PANEL_EXTENSION);
         if (component) {
             bus_component_set_restart (component, restart);
         }
-        if (component == NULL ||
+        if (component != NULL &&
             !bus_component_start (component, g_verbose)) {
             g_printerr ("Can not execute default panel program\n");
             exit (-1);
         }
-    } else if (g_strcmp0 (panel_extension, "disable") != 0 &&
-               g_strcmp0 (panel_extension, "") != 0) {
-        if (!execute_cmdline (panel_extension))
+    } else if (g_strcmp0 (emoji_extension, "disable") != 0 &&
+               g_strcmp0 (emoji_extension, "") != 0) {
+        if (!execute_cmdline (emoji_extension))
             exit (-1);
     }
 #endif
@@ -311,6 +312,47 @@ main (gint argc, gchar **argv)
             exit (-1);
     }
 
+    if (!daemonize) {
+        if (getppid () == 1) {
+            g_warning ("The parent process died.");
+            exit (0);
+        }
+#ifdef HAVE_SYS_PRCTL_H
+       /* Currently ibus-x11 detects XIOError and assume the error as the
+        * desktop session is closed and ibus-x11 calls Exit D-Bus method to
+        * exit ibus-daemon. But a few desktop sessions cause XError before
+        * XIOError and GTK does not allow to bind XError by applications and
+        * GTK calls gdk_x_error() with XError.
+        *
+        * E.g. GdkX11Screen calls XGetSelectionOwner() for "_XSETTINGS_S?"
+        * atoms during the logout but the selection owner already becomes
+        * NULL and the NULL window causes XError with
+        * gdk_x11_window_foreign_new_for_display().
+        *
+        * Since ibus-x11 exits with XError before XIOError, gnome-shell
+        * can detects the exit of ibus-daemon a little earlier and
+        * gnome-shell restarts ibus-daemon but gnome-shell dies soon.
+        * Then gnome-shell dies but ibus-daemon is alive, it's a problem.
+        * Because it causes double ibus-x11 of GDM and a login user
+        * and double XSetSelectionOwner() is not allowed for the unique
+        * "ibus" atom and the user cannot use XIM but not GtkIMModule.
+        *
+        * Probably we could fix the ibus process problem if we would fix
+        * XError about the X selection owner or stop to restart ibus-daemon
+        * in gonme-shell when the session is logging out.
+        * Maybe using SessionManager.LogoutRemote() or
+        * global.screen.get_display().get_xdisplay()
+        * But I assume there are other scenarios to cause the problem.
+        *
+        * And I decided ibus-daemon always exits with the parent's death here
+        * to avoid unexpected ibus restarts during the logout.
+        */
+        if (prctl (PR_SET_PDEATHSIG, SIGUSR1))
+            g_printerr ("Cannot bind SIGUSR1 for parent death\n");
+        else
+            signal (SIGUSR1, _sig_usr1_handler);
+#endif
+    }
     bus_server_run ();
     return 0;
 }
