@@ -2,8 +2,8 @@
 /* vim:set et sts=4: */
 /* ibus - The Input Bus
  * Copyright (C) 2008-2013 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2015-2020 Takao Fujiwara <takao.fujiwara1@gmail.com>
- * Copyright (C) 2008-2020 Red Hat, Inc.
+ * Copyright (C) 2015-2022 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2008-2022 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -110,13 +110,15 @@ static guint    _signal_preedit_end_id = 0;
 static guint    _signal_delete_surrounding_id = 0;
 static guint    _signal_retrieve_surrounding_id = 0;
 
-#if !GTK_CHECK_VERSION (3, 98, 4)
+#if GTK_CHECK_VERSION (3, 98, 4)
+static char _use_sync_mode = 2;
+#else
 static const gchar *_no_snooper_apps = NO_SNOOPER_APPS;
 static gboolean _use_key_snooper = ENABLE_SNOOPER;
 static guint    _key_snooper_id = 0;
-#endif
 
-static gboolean _use_sync_mode = FALSE;
+static char _use_sync_mode = 0;
+#endif
 
 static const gchar *_discard_password_apps  = "";
 static gboolean _use_discard_password = FALSE;
@@ -165,11 +167,19 @@ static void     ibus_im_context_set_cursor_location
 static void     ibus_im_context_set_use_preedit
                                             (GtkIMContext           *context,
                                              gboolean               use_preedit);
+#if !GTK_CHECK_VERSION (4, 1, 2)
 static void     ibus_im_context_set_surrounding
                                             (GtkIMContext  *slave,
                                              const gchar   *text,
-                                             gint           len,
-                                             gint           cursor_index);
+                                             int            len,
+                                             int            cursor_index);
+#endif
+static void     ibus_im_context_set_surrounding_with_selection
+                                            (GtkIMContext  *slave,
+                                             const gchar   *text,
+                                             int            len,
+                                             int            cursor_index,
+                                             int            anchor_index);
 
 /* static methods*/
 static void     _ibus_context_update_preedit_text_cb
@@ -356,17 +366,24 @@ ibus_im_context_commit_event (IBusIMContext *ibusimcontext,
         g_signal_emit (ibusimcontext, _signal_commit_id, 0, text->text);
         g_object_unref (text);
         _request_surrounding_text (ibusimcontext);
+#if !GTK_CHECK_VERSION (3, 98, 4)
+        /* Avoid a loop with _ibus_context_forward_key_event_cb() */
+        event->state |= IBUS_HANDLED_MASK;
+#endif
         return TRUE;
     }
    return FALSE;
 }
 
-struct _ProcessKeyEventData {
+typedef struct {
     GdkEvent *event;
     IBusIMContext *ibusimcontext;
-};
+} ProcessKeyEventData;
 
-typedef struct _ProcessKeyEventData ProcessKeyEventData;
+typedef struct {
+    GMainLoop *loop;
+    gboolean    retval;
+} ProcessKeyEventReplyData;
 
 static void
 _process_key_event_done (GObject      *object,
@@ -377,14 +394,16 @@ _process_key_event_done (GObject      *object,
 
     ProcessKeyEventData *data = (ProcessKeyEventData *)user_data;
     GdkEvent *event = data->event;
+#if GTK_CHECK_VERSION (3, 98, 4)
     IBusIMContext *ibusimcontext = data->ibusimcontext;
+#endif
     GError *error = NULL;
+    gboolean retval;
 
     g_slice_free (ProcessKeyEventData, data);
-    gboolean retval = ibus_input_context_process_key_event_async_finish (
-            context,
-            res,
-            &error);
+    retval = ibus_input_context_process_key_event_async_finish (context,
+                                                                res,
+                                                                &error);
 
     if (error != NULL) {
         g_warning ("Process Key Event failed: %s.", error->message);
@@ -415,6 +434,27 @@ _process_key_event_done (GObject      *object,
 #endif
 }
 
+static void
+_process_key_event_reply_done (GObject      *object,
+                               GAsyncResult *res,
+                               gpointer      user_data)
+{
+    IBusInputContext *context = (IBusInputContext *)object;
+    ProcessKeyEventReplyData *data = (ProcessKeyEventReplyData *)user_data;
+    GError *error = NULL;
+    gboolean retval = ibus_input_context_process_key_event_async_finish (
+            context,
+            res,
+            &error);
+    if (error != NULL) {
+        g_warning ("Process Key Event failed: %s.", error->message);
+        g_error_free (error);
+    }
+    g_return_if_fail (data);
+    data->retval = retval;
+    g_main_loop_quit (data->loop);
+}
+
 static gboolean
 _process_key_event (IBusInputContext *context,
 #if GTK_CHECK_VERSION (3, 98, 4)
@@ -446,13 +486,45 @@ _process_key_event (IBusInputContext *context,
 #endif
     keycode = hardware_keycode;
 
-    if (_use_sync_mode) {
+    switch (_use_sync_mode) {
+    case 1: {
         retval = ibus_input_context_process_key_event (context,
+                                                       keyval,
+                                                       keycode - 8,
+                                                       state);
+        break;
+    }
+    case 2: {
+        GMainLoop *loop = g_main_loop_new (NULL, TRUE);
+        ProcessKeyEventReplyData *data = NULL;
+
+        if (loop)
+            data = g_slice_new0 (ProcessKeyEventReplyData);
+        if (!data) {
+            g_warning ("Cannot wait for the reply of the process key event.");
+            retval = ibus_input_context_process_key_event (context,
+                                                           keyval,
+                                                           keycode - 8,
+                                                           state);
+            if (loop)
+                g_main_loop_quit (loop);
+            break;
+        }
+        data->loop = loop;
+        ibus_input_context_process_key_event_async (context,
             keyval,
             keycode - 8,
-            state);
+            state,
+            -1,
+            NULL,
+            _process_key_event_reply_done,
+            data);
+        g_main_loop_run (loop);
+        retval = data->retval;
+        g_slice_free (ProcessKeyEventReplyData, data);
+        break;
     }
-    else {
+    default: {
         ProcessKeyEventData *data = g_slice_new0 (ProcessKeyEventData);
 #if GTK_CHECK_VERSION (3, 98, 4)
         data->event = gdk_event_ref (event);
@@ -471,6 +543,7 @@ _process_key_event (IBusInputContext *context,
 
         retval = TRUE;
     }
+    }
 
     /* GTK4 does not provide gtk_key_snooper_install() and also
      * GtkIMContextClass->filter_keypress() cannot send the updated
@@ -631,12 +704,15 @@ _key_snooper_cb (GtkWidget   *widget,
 
     } while (0);
 
-    if (ibusimcontext != NULL) {
+    if (ibusimcontext != NULL && event->type == GDK_KEY_PRESS) {
         /* "retrieve-surrounding" signal sometimes calls unref by
          * gtk_im_multicontext_get_slave() because priv->context_id is not
          * the latest than global_context_id in GtkIMMulticontext.
          * Since _focus_im_context is gotten by the focus_in event,
          * it would be good to call ref here.
+         *
+         * Most release key events would be redundant from
+         * _ibus_context_forward_key_event_cb ().
          */
         g_object_ref (ibusimcontext);
         _request_surrounding_text (ibusimcontext);
@@ -645,7 +721,7 @@ _key_snooper_cb (GtkWidget   *widget,
 
     retval = _process_key_event (ibuscontext, event, ibusimcontext);
 
-    if (ibusimcontext != NULL) {
+    if (ibusimcontext != NULL && event->type == GDK_KEY_PRESS) {
         /* unref ibusimcontext could call ibus_im_context_finalize here
          * because "retrieve-surrounding" signal could call unref.
          */
@@ -657,24 +733,47 @@ _key_snooper_cb (GtkWidget   *widget,
 #endif
 
 static gboolean
-_get_boolean_env(const gchar *name,
-                 gboolean     defval)
+_get_boolean_env (const gchar *name,
+                  gboolean     defval)
 {
     const gchar *value = g_getenv (name);
 
     if (value == NULL)
-      return defval;
+        return defval;
 
     if (g_strcmp0 (value, "") == 0 ||
         g_strcmp0 (value, "0") == 0 ||
         g_strcmp0 (value, "false") == 0 ||
         g_strcmp0 (value, "False") == 0 ||
-        g_strcmp0 (value, "FALSE") == 0)
-      return FALSE;
+        g_strcmp0 (value, "FALSE") == 0) {
+        return FALSE;
+    }
 
     return TRUE;
 }
 
+static char
+_get_char_env (const gchar *name,
+               char         defval)
+{
+    const gchar *value = g_getenv (name);
+
+    if (value == NULL)
+        return defval;
+
+    if (g_strcmp0 (value, "") == 0 ||
+        g_strcmp0 (value, "0") == 0 ||
+        g_strcmp0 (value, "false") == 0 ||
+        g_strcmp0 (value, "False") == 0 ||
+        g_strcmp0 (value, "FALSE") == 0) {
+        return 0;
+    } else if (!g_strcmp0 (value, "2")) {
+        return 2;
+    }
+
+    return 1;
+}
+
 static void
 daemon_name_appeared (GDBusConnection *connection,
                       const gchar     *name,
@@ -724,7 +823,12 @@ ibus_im_context_class_init (IBusIMContextClass *class)
 #endif
     im_context_class->set_cursor_location = ibus_im_context_set_cursor_location;
     im_context_class->set_use_preedit = ibus_im_context_set_use_preedit;
+#if GTK_CHECK_VERSION (4, 1, 2)
+    im_context_class->set_surrounding_with_selection
+            = ibus_im_context_set_surrounding_with_selection;
+#else
     im_context_class->set_surrounding = ibus_im_context_set_surrounding;
+#endif
     gobject_class->notify = ibus_im_context_notify;
     gobject_class->finalize = ibus_im_context_finalize;
 
@@ -752,11 +856,13 @@ ibus_im_context_class_init (IBusIMContextClass *class)
         g_signal_lookup ("retrieve-surrounding", G_TYPE_FROM_CLASS (class));
     g_assert (_signal_retrieve_surrounding_id != 0);
 
-#if !GTK_CHECK_VERSION (3, 98, 4)
+#if GTK_CHECK_VERSION (3, 98, 4)
+    _use_sync_mode = _get_char_env ("IBUS_ENABLE_SYNC_MODE", 2);
+#else
     _use_key_snooper = !_get_boolean_env ("IBUS_DISABLE_SNOOPER",
                                           !(ENABLE_SNOOPER));
+    _use_sync_mode = (char)_get_char_env ("IBUS_ENABLE_SYNC_MODE", 0);
 #endif
-    _use_sync_mode = _get_boolean_env ("IBUS_ENABLE_SYNC_MODE", FALSE);
     _use_discard_password = _get_boolean_env ("IBUS_DISCARD_PASSWORD", FALSE);
 
 #define CHECK_APP_IN_CSV_ENV_VARIABLES(retval,                          \
@@ -848,33 +954,6 @@ ibus_im_context_class_fini (IBusIMContextClass *class)
     g_bus_unwatch_name (_daemon_name_watch_id);
 }
 
-/* Copied from gtk+2.0-2.20.1/modules/input/imcedilla.c to fix crosbug.com/11421.
- * Overwrite the original Gtk+'s compose table in gtk+-2.x.y/gtk/gtkimcontextsimple.c. */
-
-/* The difference between this and the default input method is the handling
- * of C+acute - this method produces C WITH CEDILLA rather than C WITH ACUTE.
- * For languages that use CCedilla and not acute, this is the preferred mapping,
- * and is particularly important for pt_BR, where the us-intl keyboard is
- * used extensively.
- */
-static guint16 cedilla_compose_seqs[] = {
-#ifdef DEPRECATED_GDK_KEYSYMS
-  GDK_dead_acute,	GDK_C,	0,	0,	0,	0x00C7,	/* LATIN_CAPITAL_LETTER_C_WITH_CEDILLA */
-  GDK_dead_acute,	GDK_c,	0,	0,	0,	0x00E7,	/* LATIN_SMALL_LETTER_C_WITH_CEDILLA */
-  GDK_Multi_key,	GDK_apostrophe,	GDK_C,  0,      0,      0x00C7, /* LATIN_CAPITAL_LETTER_C_WITH_CEDILLA */
-  GDK_Multi_key,	GDK_apostrophe,	GDK_c,  0,      0,      0x00E7, /* LATIN_SMALL_LETTER_C_WITH_CEDILLA */
-  GDK_Multi_key,	GDK_C,  GDK_apostrophe,	0,      0,      0x00C7, /* LATIN_CAPITAL_LETTER_C_WITH_CEDILLA */
-  GDK_Multi_key,	GDK_c,  GDK_apostrophe,	0,      0,      0x00E7, /* LATIN_SMALL_LETTER_C_WITH_CEDILLA */
-#else
-  GDK_KEY_dead_acute,	GDK_KEY_C,	0,	0,	0,	0x00C7,	/* LATIN_CAPITAL_LETTER_C_WITH_CEDILLA */
-  GDK_KEY_dead_acute,	GDK_KEY_c,	0,	0,	0,	0x00E7,	/* LATIN_SMALL_LETTER_C_WITH_CEDILLA */
-  GDK_KEY_Multi_key,	GDK_KEY_apostrophe,	GDK_KEY_C,  0,      0,      0x00C7, /* LATIN_CAPITAL_LETTER_C_WITH_CEDILLA */
-  GDK_KEY_Multi_key,	GDK_KEY_apostrophe,	GDK_KEY_c,  0,      0,      0x00E7, /* LATIN_SMALL_LETTER_C_WITH_CEDILLA */
-  GDK_KEY_Multi_key,	GDK_KEY_C,  GDK_KEY_apostrophe,	0,      0,      0x00C7, /* LATIN_CAPITAL_LETTER_C_WITH_CEDILLA */
-  GDK_KEY_Multi_key,	GDK_KEY_c,  GDK_KEY_apostrophe,	0,      0,      0x00E7, /* LATIN_SMALL_LETTER_C_WITH_CEDILLA */
-#endif
-};
-
 static void
 ibus_im_context_init (GObject *obj)
 {
@@ -905,15 +984,13 @@ ibus_im_context_init (GObject *obj)
 #else
     ibusimcontext->caps = IBUS_CAP_PREEDIT_TEXT | IBUS_CAP_FOCUS;
 #endif
+    if (_use_sync_mode != 1)
+        ibusimcontext->caps |= IBUS_CAP_SYNC_PROCESS_KEY;
 
     ibusimcontext->events_queue = g_queue_new ();
 
     // Create slave im context
     ibusimcontext->slave = gtk_im_context_simple_new ();
-    gtk_im_context_simple_add_table (GTK_IM_CONTEXT_SIMPLE (ibusimcontext->slave),
-                                     cedilla_compose_seqs,
-                                     4,
-                                     G_N_ELEMENTS (cedilla_compose_seqs) / (4 + 2));
 
     g_signal_connect (ibusimcontext->slave,
                       "commit",
@@ -1251,6 +1328,8 @@ ibus_im_context_reset (GtkIMContext *context)
          * IBus uses button-press-event instead until GTK is fixed.
          * https://gitlab.gnome.org/GNOME/gtk/issues/1534
          */
+        if (_use_sync_mode == 1)
+            ibus_im_context_clear_preedit_text (ibusimcontext);
         ibus_input_context_reset (ibusimcontext->ibuscontext);
     }
     gtk_im_context_reset (ibusimcontext->slave);
@@ -1364,7 +1443,7 @@ ibus_im_context_set_client_window (GtkIMContext *context,
 
     if (ibusimcontext->client_window) {
 #if !GTK_CHECK_VERSION (3, 98, 4)
-        if (ibusimcontext->use_button_press_event)
+        if (ibusimcontext->use_button_press_event && _use_sync_mode != 1)
             _connect_button_press_event (ibusimcontext, FALSE);
 #endif
         g_object_unref (ibusimcontext->client_window);
@@ -1374,7 +1453,7 @@ ibus_im_context_set_client_window (GtkIMContext *context,
     if (client != NULL) {
         ibusimcontext->client_window = g_object_ref (client);
 #if !GTK_CHECK_VERSION (3, 98, 4)
-        if (!ibusimcontext->use_button_press_event)
+        if (!ibusimcontext->use_button_press_event && _use_sync_mode != 1)
             _connect_button_press_event (ibusimcontext, TRUE);
 #endif
     }
@@ -1419,6 +1498,9 @@ static gboolean
 _set_cursor_location_internal (IBusIMContext *ibusimcontext)
 {
     GdkRectangle area;
+#if GTK_CHECK_VERSION (3, 98, 4)
+    GtkWidget *root;
+#endif
 
     if(ibusimcontext->client_window == NULL ||
        ibusimcontext->ibuscontext == NULL) {
@@ -1427,8 +1509,27 @@ _set_cursor_location_internal (IBusIMContext *ibusimcontext)
 
     area = ibusimcontext->cursor_area;
 
-#if !GTK_CHECK_VERSION (3, 98, 4)
 #ifdef GDK_WINDOWING_WAYLAND
+#if GTK_CHECK_VERSION (3, 98, 4)
+    root = GTK_WIDGET (gtk_widget_get_root (ibusimcontext->client_window));
+     /* FIXME: GTK_STYLE_CLASS_TITLEBAR is available in GTK3 but not GTK4.
+      * gtk_css_boxes_get_content_rect() is available in GTK4 but it's an
+      * internal API and calculate the window edge 32 in GTK3.
+      */
+    area.y += 32;
+    area.width = 50; /* FIXME: Why 50 meets the cursor position? */
+    area.height = gtk_widget_get_height (root);
+    area.height += 32;
+    if (GDK_IS_WAYLAND_DISPLAY (gdk_display_get_default ())) {
+        ibus_input_context_set_cursor_location_relative (
+            ibusimcontext->ibuscontext,
+            area.x,
+            area.y,
+            area.width,
+            area.height);
+        return FALSE;
+    }
+#else
     if (GDK_IS_WAYLAND_DISPLAY (gdk_display_get_default ())) {
         gdouble px, py;
         GdkWindow *parent;
@@ -1454,23 +1555,23 @@ _set_cursor_location_internal (IBusIMContext *ibusimcontext)
 #endif
 #endif
 
-    if (area.x == -1 && area.y == -1 && area.width == 0 && area.height == 0) {
 #if GTK_CHECK_VERSION (3, 98, 4)
-        area.x = 0;
-        area.y += gtk_widget_get_height (ibusimcontext->client_window);
 #elif GTK_CHECK_VERSION (2, 91, 0)
+    if (area.x == -1 && area.y == -1 && area.width == 0 && area.height == 0) {
         area.x = 0;
         area.y += gdk_window_get_height (ibusimcontext->client_window);
+    }
 #else
+    if (area.x == -1 && area.y == -1 && area.width == 0 && area.height == 0) {
         gint w, h;
         gdk_drawable_get_size (ibusimcontext->client_window, &w, &h);
         area.y += h;
         area.x = 0;
-#endif
     }
+#endif
 
 #if GTK_CHECK_VERSION (3, 98, 4)
-#ifdef GDK_WINDOWING_X11
+#if defined(GDK_WINDOWING_X11)
     GdkDisplay *display = gtk_widget_get_display (ibusimcontext->client_window);
     if (GDK_IS_X11_DISPLAY (display)) {
         Display *xdisplay = gdk_x11_display_get_xdisplay (display);
@@ -1490,21 +1591,10 @@ _set_cursor_location_internal (IBusIMContext *ibusimcontext)
         XGetWindowAttributes (xdisplay, window, &xwa);
         area.x = x - xwa.x + area.x;
         area.y = y - xwa.y + area.y;
-        area.width = xwa.width;
+        area.width = 50; /* FIXME: Why 50 meets the cursor position? */
         area.height = xwa.height;
     }
 #endif
-#elif GTK_CHECK_VERSION (3, 93, 0)
-    {
-        GtkNative *native = gtk_widget_get_native (
-                ibusimcontext->client_window);
-        GdkSurface *surface = gtk_native_get_surface (native);
-        int root_x = 0;
-        int root_y = 0;
-        gdk_surface_get_position (surface, &root_x, &root_y);
-        area.x += root_x;
-        area.y += root_y;
-    }
 #else
     gdk_window_get_root_coords (ibusimcontext->client_window,
                                 area.x, area.y,
@@ -1526,12 +1616,17 @@ ibus_im_context_set_cursor_location (GtkIMContext *context, GdkRectangle *area)
 
     IBusIMContext *ibusimcontext = IBUS_IM_CONTEXT (context);
 
+#if !GTK_CHECK_VERSION (3, 93, 0)
+    /* The area is the relative coordinates and this has to get the absolute
+     * ones in _set_cursor_location_internal() since GTK 4.0.
+     */
     if (ibusimcontext->cursor_area.x == area->x &&
         ibusimcontext->cursor_area.y == area->y &&
         ibusimcontext->cursor_area.width == area->width &&
         ibusimcontext->cursor_area.height == area->height) {
         return;
     }
+#endif
     ibusimcontext->cursor_area = *area;
     _set_cursor_location_internal (ibusimcontext);
     gtk_im_context_set_cursor_location (ibusimcontext->slave, area);
@@ -1621,11 +1716,27 @@ get_selection_anchor_point (IBusIMContext *ibusimcontext,
     return anchor;
 }
 
+#if !GTK_CHECK_VERSION (4, 1, 2)
 static void
 ibus_im_context_set_surrounding (GtkIMContext  *context,
                                  const gchar   *text,
-                                 gint           len,
-                                 gint           cursor_index)
+                                 int            len,
+                                 int            cursor_index)
+{
+    ibus_im_context_set_surrounding_with_selection (context,
+                                                    text,
+                                                    len,
+                                                    cursor_index,
+                                                    cursor_index);
+}
+#endif
+
+static void
+ibus_im_context_set_surrounding_with_selection (GtkIMContext  *context,
+                                                const gchar   *text,
+                                                int            len,
+                                                int            cursor_index,
+                                                int            anchor_index)
 {
     g_return_if_fail (context != NULL);
     g_return_if_fail (IBUS_IS_IM_CONTEXT (context));
@@ -1647,18 +1758,26 @@ ibus_im_context_set_surrounding (GtkIMContext  *context,
         ibustext = ibus_text_new_from_string (p);
         g_free (p);
 
-        guint anchor_pos = get_selection_anchor_point (ibusimcontext,
-                                                       cursor_pos,
-                                                       utf8_len);
+        gint anchor_pos = get_selection_anchor_point (ibusimcontext,
+                                                      cursor_pos,
+                                                      utf8_len);
         ibus_input_context_set_surrounding_text (ibusimcontext->ibuscontext,
                                                  ibustext,
                                                  cursor_pos,
                                                  anchor_pos);
     }
+#if GTK_CHECK_VERSION (4, 1, 2)
+    gtk_im_context_set_surrounding_with_selection (ibusimcontext->slave,
+                                                   text,
+                                                   len,
+                                                   cursor_index,
+                                                   anchor_index);
+#else
     gtk_im_context_set_surrounding (ibusimcontext->slave,
                                     text,
                                     len,
                                     cursor_index);
+#endif
 }
 
 static void
@@ -1816,7 +1935,11 @@ _create_gdk_event (IBusIMContext *ibusimcontext,
         if (event->state & GDK_CONTROL_MASK) {
             if ((c >= '@' && c < '\177') || c == ' ') c &= 0x1F;
             else if (c == '2') {
+#if GLIB_CHECK_VERSION (2, 68, 0)
+                event->string = g_memdup2 ("\0\0", 2);
+#else
                 event->string = g_memdup ("\0\0", 2);
+#endif
                 event->length = 1;
                 buf[0] = '\0';
                 goto out;
@@ -1873,14 +1996,20 @@ _ibus_context_forward_key_event_cb (IBusInputContext  *ibuscontext,
 #if GTK_CHECK_VERSION (3, 98, 4)
     int group = 0;
     g_return_if_fail (GTK_IS_IM_CONTEXT (ibusimcontext));
-    if (keycode == 0 && ibusimcontext->client_window) {
-        GdkDisplay *display = gtk_widget_get_display (ibusimcontext->client_window);
+    if (keycode != 0) {
+        keycode += 8; // to GTK keycode
+    } else if (ibusimcontext->client_window) {
+        GdkDisplay *display =
+                gtk_widget_get_display (ibusimcontext->client_window);
         GdkKeymapKey *keys = NULL;
         gint n_keys = 0;
-        if (!gdk_display_map_keyval (display, keyval, &keys, &n_keys))
+        if (gdk_display_map_keyval (display, keyval, &keys, &n_keys)) {
+            keycode = keys->keycode;
+            group = keys->group;
+            g_free (keys);
+        } else {
             g_warning ("Failed to parse keycode from keyval %x", keyval);
-       keycode = keys->keycode;
-       group = keys->group;
+        }
     }
     gtk_im_context_filter_key (
         GTK_IM_CONTEXT (ibusimcontext),
@@ -1892,6 +2021,20 @@ _ibus_context_forward_key_event_cb (IBusInputContext  *ibuscontext,
         (GdkModifierType)state,
         group);
 #else
+    if (keycode == 0 && ibusimcontext->client_window) {
+        GdkDisplay *display =
+                gdk_window_get_display (ibusimcontext->client_window);
+        GdkKeymap *keymap = gdk_keymap_get_for_display (display);
+        GdkKeymapKey *keys = NULL;
+        gint n_keys = 0;
+        if (gdk_keymap_get_entries_for_keyval (keymap, keyval, &keys, &n_keys))
+            keycode = keys->keycode;
+        else
+            g_warning ("Failed to parse keycode from keyval %x", keyval);
+        /* _create_gdk_event() will add 8 to keycode. */
+        if (keycode != 0)
+            keycode -= 8;
+    }
     GdkEventKey *event = _create_gdk_event (ibusimcontext, keyval, keycode, state);
     gdk_event_put ((GdkEvent *)event);
     gdk_event_free ((GdkEvent *)event);
@@ -1931,7 +2074,8 @@ _ibus_context_update_preedit_text_cb (IBusInputContext  *ibuscontext,
 
 #if !GTK_CHECK_VERSION (3, 98, 4)
     if (!ibusimcontext->use_button_press_event &&
-        mode == IBUS_ENGINE_PREEDIT_COMMIT) {
+        mode == IBUS_ENGINE_PREEDIT_COMMIT &&
+        _use_sync_mode != 1) {
         if (ibusimcontext->client_window) {
             _connect_button_press_event (ibusimcontext, TRUE);
         }
@@ -2138,6 +2282,8 @@ _create_input_context_done (IBusBus       *bus,
 static void
 _create_input_context (IBusIMContext *ibusimcontext)
 {
+    gchar *prgname = g_strdup (g_get_prgname());
+    gchar *client_name;
     IDEBUG ("%s", __FUNCTION__);
 
     g_assert (ibusimcontext->ibuscontext == NULL);
@@ -2146,11 +2292,24 @@ _create_input_context (IBusIMContext *ibusimcontext)
 
     ibusimcontext->cancellable = g_cancellable_new ();
 
+    if (!prgname)
+        prgname = g_strdup_printf ("(%d)", getpid ());
+    client_name = g_strdup_printf ("%s:%s",
+#if GTK_CHECK_VERSION (3, 98, 4)
+                                   "gtk4-im",
+#elif GTK_CHECK_VERSION (2, 91, 0)
+                                   "gtk3-im",
+#else
+                                   "gtk-im",
+#endif
+                                   prgname);
+    g_free (prgname);
     ibus_bus_create_input_context_async (_bus,
-            "gtk-im", -1,
+            client_name, -1,
             ibusimcontext->cancellable,
             (GAsyncReadyCallback)_create_input_context_done,
             g_object_ref (ibusimcontext));
+    g_free (client_name);
 }
 
 /* Callback functions for slave context */
@@ -2267,6 +2426,8 @@ _create_fake_input_context_done (IBusBus       *bus,
                       NULL);
 
     guint32 caps = IBUS_CAP_PREEDIT_TEXT | IBUS_CAP_FOCUS | IBUS_CAP_SURROUNDING_TEXT;
+    if (_use_sync_mode != 1)
+        caps |= IBUS_CAP_SYNC_PROCESS_KEY;
     ibus_input_context_set_capabilities (_fake_context, caps);
 
     /* focus in/out the fake context */
