@@ -45,6 +45,12 @@ for schema in $schemas; do
   done
 done
 
+# dbus-launch and gsettings run /usr/lib*/gvfsd-fuse $TMPDIR/cache/gvfs -f
+# via systemd since gvfs 1.45.90 in Fedora 33
+# and rm $TMPDIR could be failed until umount would be called.
+if [ -d $TMPDIR/cache/gvfs ] ; then
+    umount $TMPDIR/cache/gvfs
+fi
 rm -rf $TMPDIR
 
 kill $DBUS_SESSION_BUS_PID
