@@ -26,6 +26,12 @@
 
 G_BEGIN_DECLS
 
+/* Checks if a keysym is a dead key. Dead key keysym values are defined in
+ * ibuskeysyms.h and the first is GDK_KEY_dead_grave.
+ */
+#define IS_DEAD_KEY(k) \
+      ((k) >= IBUS_KEY_dead_grave && (k) <= IBUS_KEY_dead_greek)
+
 extern const IBusComposeTableCompactEx ibus_compose_table_compact;
 extern const IBusComposeTableCompactEx ibus_compose_table_compact_32bit;
 
@@ -37,15 +43,31 @@ struct _IBusComposeTablePrivate
     gsize second_size;
 };
 
+struct _IBusComposeTableCompactPrivate
+{
+    const guint32 *data2;
+};
+
 gboolean ibus_check_algorithmically (const guint16              *compose_buffer,
-                                     gint                        n_compose,
+                                     int                         n_compose,
                                      gunichar                   *output);
-gboolean ibus_check_compact_table   (const IBusComposeTableCompactEx
+gboolean ibus_compose_table_check   (const IBusComposeTableEx   *table,
+                                     guint16                    *compose_buffer,
+                                     int                         n_compose,
+                                     gboolean                   *compose_finish,
+                                     gboolean                   *compose_match,
+                                     GString                    *output,
+                                     gboolean                    is_32bit);
+gboolean ibus_compose_table_compact_check
+                                    (const IBusComposeTableCompactEx
                                                                 *table,
                                      guint16                    *compose_buffer,
-                                     gint                        n_compose,
+                                     int                         n_compose,
                                      gboolean                   *compose_finish,
                                      gunichar                  **output_chars);
+gunichar ibus_keysym_to_unicode     (guint16                     keysym,
+                                     gboolean                    combining,
+                                     gboolean                   *need_space);
 
 G_END_DECLS
 
