@@ -108,6 +108,10 @@ typedef enum
  * @IBUS_CAP_PROPERTY: UI is capable to have property.
  * @IBUS_CAP_SURROUNDING_TEXT: Client can provide surround text,
  *  or IME can handle surround text.
+ * @IBUS_CAP_OSK: UI is owned by on-screen keyboard.
+ * @IBUS_CAP_SYNC_PROCESS_KEY: Asynchronous process key events are not
+ *  supported and the ibus_engine_forward_key_event() should not be
+ *  used for the return value of #IBusEngine::process_key_event().
  *
  * Capability flags of UI.
  */
@@ -118,6 +122,8 @@ typedef enum {
     IBUS_CAP_FOCUS              = 1 << 3,
     IBUS_CAP_PROPERTY           = 1 << 4,
     IBUS_CAP_SURROUNDING_TEXT   = 1 << 5,
+    IBUS_CAP_OSK                = 1 << 6,
+    IBUS_CAP_SYNC_PROCESS_KEY   = 1 << 7,
 } IBusCapabilite;
 
 /**
