@@ -6,6 +6,7 @@
 #define GREEN "\033[0;32m"
 #define RED   "\033[0;31m"
 #define NC    "\033[0m"
+#define X11_DATADIR X11_DATA_PREFIX "/share/X11/locale"
 
 IBusBus *m_bus;
 gchar *m_compose_file;
@@ -35,7 +36,7 @@ get_compose_path ()
             break;
         if (g_strcmp0 (*l, "C") == 0)
             break;
-        compose_path = g_build_filename ("/usr/share/X11/locale",
+        compose_path = g_build_filename (X11_DATADIR,
                                          *l,
                                          "Compose",
                                          NULL);
@@ -155,7 +156,7 @@ set_engine_cb (GObject *object, GAsyncResult *res, gpointer data)
     for (i = 0;
          i < (m_compose_table->n_seqs * index_stride);
          i += index_stride) {
-        for (j = i; j < i + (index_stride - 1); j++) {
+        for (j = i; j < i + (index_stride - 2); j++) {
             guint keyval = m_compose_table->data[j];
             guint keycode = 0;
             guint modifiers = 0;
@@ -175,7 +176,7 @@ set_engine_cb (GObject *object, GAsyncResult *res, gpointer data)
         for (i = 0;
              i < (priv->first_n_seqs * index_stride);
              i += index_stride) {
-            for (j = i; j < i + (index_stride - 1); j++) {
+            for (j = i; j < i + (index_stride - 2); j++) {
                 guint keyval = priv->data_first[j];
                 guint keycode = 0;
                 guint modifiers = 0;
@@ -360,11 +361,13 @@ main (int argc, char *argv[])
     /* Avoid a warning of "AT-SPI: Could not obtain desktop path or name"
      * with gtk_main().
      */
-    g_setenv ("NO_AT_BRIDGE", "1", TRUE);
+    if (!g_setenv ("NO_AT_BRIDGE", "1", TRUE))
+        g_message ("Failed setenv NO_AT_BRIDGE\n");
     g_test_init (&argc, &argv, NULL);
     gtk_init (&argc, &argv);
 
-    m_srcdir = argc > 1 ? g_strdup (argv[1]) : g_strdup (".");
+    m_srcdir = (argc > 1 && strlen (argv[1]) < FILENAME_MAX)
+            ? g_strdup (argv[1]) : g_strdup (".");
     m_compose_file = g_strdup (g_getenv ("COMPOSE_FILE"));
 #if GLIB_CHECK_VERSION (2, 58, 0)
     test_name = g_get_language_names_with_category ("LC_CTYPE")[0];
